/*
 * Copyright (c) 2012-2023, Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <WebCore/LayoutUnit.h>

using namespace WebCore;

namespace TestWebKitAPI {

namespace {
constexpr float tolerance = 1.0f / kFixedPointDenominator;
}

TEST(WebCoreLayoutUnit, LayoutUnitInt)
{
    ASSERT_EQ(LayoutUnit(INT_MIN).toInt(), intMinForLayoutUnit);
    ASSERT_EQ(LayoutUnit(INT_MIN / 2).toInt(), intMinForLayoutUnit);
    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit - 1).toInt(), intMinForLayoutUnit); 
    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit).toInt(), intMinForLayoutUnit);
    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit + 1).toInt(), intMinForLayoutUnit + 1);
    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit / 2).toInt(), intMinForLayoutUnit / 2);
    ASSERT_EQ(LayoutUnit(-10000).toInt(), -10000);
    ASSERT_EQ(LayoutUnit(-1000).toInt(), -1000);
    ASSERT_EQ(LayoutUnit(-100).toInt(), -100);
    ASSERT_EQ(LayoutUnit(-10).toInt(), -10);
    ASSERT_EQ(LayoutUnit(-1).toInt(), -1);
    ASSERT_EQ(LayoutUnit(0).toInt(), 0);
    ASSERT_EQ(LayoutUnit(1).toInt(), 1);
    ASSERT_EQ(LayoutUnit(100).toInt(), 100);
    ASSERT_EQ(LayoutUnit(1000).toInt(), 1000);
    ASSERT_EQ(LayoutUnit(10000).toInt(), 10000);
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit / 2).toInt(), intMaxForLayoutUnit / 2);
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit - 1).toInt(), intMaxForLayoutUnit - 1);
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit).toInt(), intMaxForLayoutUnit);
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit + 1).toInt(), intMaxForLayoutUnit);
    ASSERT_EQ(LayoutUnit(INT_MAX / 2).toInt(), intMaxForLayoutUnit);
    ASSERT_EQ(LayoutUnit(INT_MAX).toInt(), intMaxForLayoutUnit);
}

TEST(WebCoreLayoutUnit, LayoutUnitFloat)
{
    ASSERT_FLOAT_EQ(LayoutUnit(1.0f).toFloat(), 1.0f);
    ASSERT_FLOAT_EQ(LayoutUnit(1.25f).toFloat(), 1.25f);
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit(1.25f + tolerance / 2));
    ASSERT_EQ(LayoutUnit(-2.0f), LayoutUnit(-2.0f - tolerance / 2));
    ASSERT_NEAR(LayoutUnit(1.1f).toFloat(), 1.1f, tolerance);
    ASSERT_NEAR(LayoutUnit(1.33f).toFloat(), 1.33f, tolerance);
    ASSERT_NEAR(LayoutUnit(1.3333f).toFloat(), 1.3333f, tolerance);
    ASSERT_NEAR(LayoutUnit(1.53434f).toFloat(), 1.53434f, tolerance);
    ASSERT_NEAR(LayoutUnit(345634).toFloat(), 345634.0f, tolerance);
    ASSERT_NEAR(LayoutUnit(345634.12335f).toFloat(), 345634.12335f, tolerance);
    ASSERT_NEAR(LayoutUnit(-345634.12335f).toFloat(), -345634.12335f, tolerance);
    ASSERT_NEAR(LayoutUnit(-345634).toFloat(), -345634.0f, tolerance);    
    ASSERT_NEAR(LayoutUnit(33554432.f).toFloat(), 33554432.f, tolerance);
    ASSERT_NEAR(LayoutUnit(-33554432.f).toFloat(), -33554432.f, tolerance);
    ASSERT_NEAR(LayoutUnit(33554432.f).toDouble(), 33554432.f, tolerance);
    ASSERT_NEAR(LayoutUnit(-33554432.f).toDouble(), -33554432.f, tolerance);
}

TEST(WebCoreLayoutUnit, LayoutUnitRounding)
{
    ASSERT_EQ(LayoutUnit(-1.9f).round(), -2);
    ASSERT_EQ(LayoutUnit(-1.6f).round(), -2);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-1.51f).round(), -2);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-1.5f).round(), -1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-1.49f).round(), -1);
    ASSERT_EQ(LayoutUnit(-1.0f).round(), -1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-0.99f).round(), -1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-0.51f).round(), -1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-0.50f).round(), 0);
    ASSERT_EQ(LayoutUnit::fromFloatRound(-0.49f).round(), 0);
    ASSERT_EQ(LayoutUnit(-0.1f).round(), 0);
    ASSERT_EQ(LayoutUnit(0.0f).round(), 0);
    ASSERT_EQ(LayoutUnit(0.1f).round(), 0);
    ASSERT_EQ(LayoutUnit::fromFloatRound(0.49f).round(), 0);
    ASSERT_EQ(LayoutUnit::fromFloatRound(0.50f).round(), 1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(0.51f).round(), 1);
    ASSERT_EQ(LayoutUnit(0.99f).round(), 1);
    ASSERT_EQ(LayoutUnit(1.0f).round(), 1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(1.49f).round(), 1);
    ASSERT_EQ(LayoutUnit::fromFloatRound(1.5f).round(), 2);
    ASSERT_EQ(LayoutUnit::fromFloatRound(1.51f).round(), 2);
    // The fractional part of LayoutUnit::Max() is 0x3f, so it should round up.
    ASSERT_EQ(((std::numeric_limits<int>::max() / kFixedPointDenominator) + 1), LayoutUnit::max().round());
    // The fractional part of LayoutUnit::Min() is 0, so the next bigger possible
    // value should round down.
    LayoutUnit epsilon;
    epsilon.setRawValue(1);
    ASSERT_EQ(((std::numeric_limits<int>::min() / kFixedPointDenominator)), (LayoutUnit::min() + epsilon).round());
}

TEST(WebCoreLayoutUnit, LayoutUnitMultiplication)
{
    ASSERT_EQ((LayoutUnit(1) * LayoutUnit(1)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(1) * LayoutUnit(2)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(2) * LayoutUnit(1)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(2) * LayoutUnit(0.5)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(0.5) * LayoutUnit(2)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(100) * LayoutUnit(1)).toInt(), 100);
    
    ASSERT_EQ((LayoutUnit(-1) * LayoutUnit(1)).toInt(), -1);
    ASSERT_EQ((LayoutUnit(-1) * LayoutUnit(2)).toInt(), -2);
    ASSERT_EQ((LayoutUnit(-2) * LayoutUnit(1)).toInt(), -2);
    ASSERT_EQ((LayoutUnit(-2) * LayoutUnit(0.5)).toInt(), -1);
    ASSERT_EQ((LayoutUnit(-0.5) * LayoutUnit(2)).toInt(), -1);
    ASSERT_EQ((LayoutUnit(-100) * LayoutUnit(1)).toInt(), -100);
    
    ASSERT_EQ((LayoutUnit(-1) * LayoutUnit(-1)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(-1) * LayoutUnit(-2)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(-2) * LayoutUnit(-1)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(-2) * LayoutUnit(-0.5)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(-0.5) * LayoutUnit(-2)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(-100) * LayoutUnit(-1)).toInt(), 100);
    
    ASSERT_EQ((LayoutUnit(100) * LayoutUnit(3.33)).round(), 333);
    ASSERT_EQ((LayoutUnit(-100) * LayoutUnit(3.33)).round(), -333);
    ASSERT_EQ((LayoutUnit(-100) * LayoutUnit(-3.33)).round(), 333);

    size_t aHundredSizeT = 100;
    ASSERT_EQ((LayoutUnit(aHundredSizeT) * LayoutUnit(1)).toInt(), 100);
    ASSERT_EQ((aHundredSizeT * LayoutUnit(4)).toInt(), 400);
    ASSERT_EQ((LayoutUnit(4) * aHundredSizeT).toInt(), 400);

    int quarterMax = intMaxForLayoutUnit / 4;
    ASSERT_EQ((LayoutUnit(quarterMax) * LayoutUnit(2)).toInt(), quarterMax * 2);
    ASSERT_EQ((LayoutUnit(quarterMax) * LayoutUnit(3)).toInt(), quarterMax * 3);
    ASSERT_EQ((LayoutUnit(quarterMax) * LayoutUnit(4)).toInt(), quarterMax * 4);
    ASSERT_EQ((LayoutUnit(quarterMax) * LayoutUnit(5)).toInt(), intMaxForLayoutUnit);

    size_t overflowIntSizeT = intMaxForLayoutUnit * 4;
    ASSERT_EQ((LayoutUnit(overflowIntSizeT) * LayoutUnit(2)).toInt(), intMaxForLayoutUnit);
    ASSERT_EQ((overflowIntSizeT * LayoutUnit(4)).toInt(), intMaxForLayoutUnit);
    ASSERT_EQ((LayoutUnit(4) * overflowIntSizeT).toInt(), intMaxForLayoutUnit);
}

TEST(WebCoreLayoutUnit, LayoutUnitDivision)
{
    ASSERT_EQ((LayoutUnit(1) / LayoutUnit(1)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(1) / LayoutUnit(2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(2) / LayoutUnit(1)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(2) / LayoutUnit(0.5)).toInt(), 4);
    ASSERT_EQ((LayoutUnit(0.5) / LayoutUnit(2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(100) / LayoutUnit(10)).toInt(), 10);
    ASSERT_FLOAT_EQ((LayoutUnit(1) / LayoutUnit(2)).toFloat(), 0.5f);
    ASSERT_FLOAT_EQ((LayoutUnit(0.5) / LayoutUnit(2)).toFloat(), 0.25f);
    
    ASSERT_EQ((LayoutUnit(-1) / LayoutUnit(1)).toInt(), -1);
    ASSERT_EQ((LayoutUnit(-1) / LayoutUnit(2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(-2) / LayoutUnit(1)).toInt(), -2);
    ASSERT_EQ((LayoutUnit(-2) / LayoutUnit(0.5)).toInt(), -4);
    ASSERT_EQ((LayoutUnit(-0.5) / LayoutUnit(2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(-100) / LayoutUnit(10)).toInt(), -10);
    ASSERT_FLOAT_EQ((LayoutUnit(-1) / LayoutUnit(2)).toFloat(), -0.5f);
    ASSERT_FLOAT_EQ((LayoutUnit(-0.5) / LayoutUnit(2)).toFloat(), -0.25f);
    
    ASSERT_EQ((LayoutUnit(-1) / LayoutUnit(-1)).toInt(), 1);
    ASSERT_EQ((LayoutUnit(-1) / LayoutUnit(-2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(-2) / LayoutUnit(-1)).toInt(), 2);
    ASSERT_EQ((LayoutUnit(-2) / LayoutUnit(-0.5)).toInt(), 4);
    ASSERT_EQ((LayoutUnit(-0.5) / LayoutUnit(-2)).toInt(), 0);
    ASSERT_EQ((LayoutUnit(-100) / LayoutUnit(-10)).toInt(), 10);
    ASSERT_FLOAT_EQ((LayoutUnit(-1) / LayoutUnit(-2)).toFloat(), 0.5f);
    ASSERT_FLOAT_EQ((LayoutUnit(-0.5) / LayoutUnit(-2)).toFloat(), 0.25f);
    
    size_t aHundredSizeT = 100;
    ASSERT_EQ((LayoutUnit(aHundredSizeT) / LayoutUnit(2)).toInt(), 50);
    ASSERT_EQ((aHundredSizeT / LayoutUnit(4)).toInt(), 25);
    ASSERT_EQ((LayoutUnit(400) / aHundredSizeT).toInt(), 4);

    ASSERT_EQ((LayoutUnit(intMaxForLayoutUnit) / LayoutUnit(2)).toInt(), intMaxForLayoutUnit / 2);
    ASSERT_EQ((LayoutUnit(intMaxForLayoutUnit) / LayoutUnit(0.5)).toInt(), intMaxForLayoutUnit);
}

TEST(WebCoreLayoutUnit, LayoutUnitCeil)
{
    ASSERT_EQ(LayoutUnit(0).ceil(), 0);
    ASSERT_EQ(LayoutUnit(0.1).ceil(), 1);
    ASSERT_EQ(LayoutUnit(0.5).ceil(), 1);
    ASSERT_EQ(LayoutUnit(0.9).ceil(), 1);
    ASSERT_EQ(LayoutUnit(1.0).ceil(), 1);
    ASSERT_EQ(LayoutUnit(1.1).ceil(), 2);
    
    ASSERT_EQ(LayoutUnit(-0.1).ceil(), 0);
    ASSERT_EQ(LayoutUnit(-0.5).ceil(), 0);
    ASSERT_EQ(LayoutUnit(-0.9).ceil(), 0);
    ASSERT_EQ(LayoutUnit(-1.0).ceil(), -1);
    
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit).ceil(), intMaxForLayoutUnit);
    ASSERT_EQ((LayoutUnit(intMaxForLayoutUnit) - LayoutUnit(0.5)).ceil(), intMaxForLayoutUnit);
    ASSERT_EQ((LayoutUnit(intMaxForLayoutUnit) - LayoutUnit(1)).ceil(), intMaxForLayoutUnit - 1);

    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit).ceil(), intMinForLayoutUnit);
}

TEST(WebCoreLayoutUnit, LayoutUnitFloor)
{
    ASSERT_EQ(LayoutUnit(0).floor(), 0);
    ASSERT_EQ(LayoutUnit(0.1).floor(), 0);
    ASSERT_EQ(LayoutUnit(0.5).floor(), 0);
    ASSERT_EQ(LayoutUnit(0.9).floor(), 0);
    ASSERT_EQ(LayoutUnit(1.0).floor(), 1);
    ASSERT_EQ(LayoutUnit(1.1).floor(), 1);
    
    ASSERT_EQ(LayoutUnit(-0.1).floor(), -1);
    ASSERT_EQ(LayoutUnit(-0.5).floor(), -1);
    ASSERT_EQ(LayoutUnit(-0.9).floor(), -1);
    ASSERT_EQ(LayoutUnit(-1.0).floor(), -1);
    
    ASSERT_EQ(LayoutUnit(intMaxForLayoutUnit).floor(), intMaxForLayoutUnit);

    ASSERT_EQ(LayoutUnit(intMinForLayoutUnit).floor(), intMinForLayoutUnit);
    ASSERT_EQ((LayoutUnit(intMinForLayoutUnit) + LayoutUnit(0.5)).floor(), intMinForLayoutUnit);
    ASSERT_EQ((LayoutUnit(intMinForLayoutUnit) + LayoutUnit(1)).floor(), intMinForLayoutUnit + 1);
}

TEST(WebCoreLayoutUnit, LayoutUnitPixelSnapping)
{
    for (int i = -100000; i <= 100000; ++i) {
        ASSERT_EQ(roundToDevicePixel(LayoutUnit(i), 1), i);
        ASSERT_EQ(roundToDevicePixel(LayoutUnit(i), 2), i);
        ASSERT_EQ(roundToDevicePixel(LayoutUnit(i), 3), i);
    }

    for (float i = -10000; i < 0; i = i + 0.5)
        ASSERT_FLOAT_EQ(roundToDevicePixel(LayoutUnit(i), 2), i);

    for (float i = -10000.25; i < 0; i = i + 0.5)
        ASSERT_FLOAT_EQ(roundToDevicePixel(LayoutUnit(i), 2), i + 0.25);
}

TEST(WebCoreLayoutUnit, LayoutUnitUnaryMinus)
{
    ASSERT_EQ(LayoutUnit(), -LayoutUnit());
    ASSERT_EQ(LayoutUnit(999), -LayoutUnit(-999));
    ASSERT_EQ(LayoutUnit(-999), -LayoutUnit(999));

    LayoutUnit negativeMax;
    negativeMax.setRawValue(LayoutUnit::min().rawValue() + 1);
    ASSERT_EQ(negativeMax, -LayoutUnit::max());
    ASSERT_EQ(LayoutUnit::max(), -negativeMax);
    // -LayoutUnit::min() is saturated to LayoutUnit::max()
    ASSERT_EQ(LayoutUnit::max(), -LayoutUnit::min());
}

TEST(WebCoreLayoutUnit, FromFloatCeil)
{
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit::fromFloatCeil(1.25f));
    ASSERT_EQ(LayoutUnit(1.25f + tolerance), LayoutUnit::fromFloatCeil(1.25f + tolerance / 2));
    ASSERT_EQ(LayoutUnit(), LayoutUnit::fromFloatCeil(-tolerance / 2));

#if !COMPILER(MSVC)
    using Limits = std::numeric_limits<float>;
    // Larger than max()
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatCeil(Limits::max()));
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatCeil(Limits::infinity()));
    // Smaller than Min()
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatCeil(Limits::lowest()));
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatCeil(-Limits::infinity()));
#endif
}

TEST(WebCoreLayoutUnit, FromFloatFloor)
{
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit::fromFloatFloor(1.25f));
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit::fromFloatFloor(1.25f + tolerance / 2));
    ASSERT_EQ(LayoutUnit(-tolerance), LayoutUnit::fromFloatFloor(-tolerance / 2));

#if !COMPILER(MSVC)
    using Limits = std::numeric_limits<float>;
    // Larger than max()
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatFloor(Limits::max()));
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatFloor(Limits::infinity()));
    // Smaller than min()
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatFloor(Limits::lowest()));
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatFloor(-Limits::infinity()));
#endif
}

TEST(WebCoreLayoutUnit, FromFloatRound)
{
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit::fromFloatRound(1.25f));
    ASSERT_EQ(LayoutUnit(1.25f), LayoutUnit::fromFloatRound(1.25f + tolerance / 4));
    ASSERT_EQ(LayoutUnit(1.25f + tolerance), LayoutUnit::fromFloatRound(1.25f + tolerance * 3 / 4));
    ASSERT_EQ(LayoutUnit(-tolerance), LayoutUnit::fromFloatRound(-tolerance * 3 / 4));

    using Limits = std::numeric_limits<float>;
    // Larger than max()
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatRound(Limits::max()));
    ASSERT_EQ(LayoutUnit::max(), LayoutUnit::fromFloatRound(Limits::infinity()));
    // Smaller than min()
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatRound(Limits::lowest()));
    ASSERT_EQ(LayoutUnit::min(), LayoutUnit::fromFloatRound(-Limits::infinity()));
}

} // namespace TestWebKitAPI
